/*
 * io.c --
 *
 *      Este fichero implementa un par de dispositivos simples de
 *      entrada/salida mapeados en memoria para el simulador DLXV.
 *
 * Copyright 1989 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:                                   *
 *      Simulador Vectorial para PC bajo entorno Windows        *
 *                                                              *
 *      Autor: Roberto Calpe Lorenzo                            *
 *                                                              *
 ****************************************************************/

#include <bwcc.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <io.h>
#include <conio.h>
#include "dlx.h"

/*
 * La entrada por el terminal es manejada viendo la entrada del terminal
 * cada cierto tiempo durante la simulacion (el simulador esta todavia
 * ocupado esperando durante la entrada por lo tanto esto no hace que las
 * cosas funcionen mal). Para la salida por el terminal, el caracter sale
 * por el terminal de forma inmediata, pero el dispositivo de salida no es
 * marcado como preparado hasta muchas instrucciones despues.
 */

/*
 * Cada cuantas instrucciones hemos de comprobar la entrada:
 */

#define INPUT_WAIT      5000L

/*
 * Cuantas instrucciones han de pasar entre que un caracter es recibido
 * para mandarlo por la salida y se considera completa la transmision.
 */

#define OUTPUT_DELAY    500L

/*
 * Declaraciones de procedimientos definidos mas tarde en este fichero:
 */

static void     CheckInput();
static void     InputTimerProc();
static void     MarkOutput();

/*
 *----------------------------------------------------------------------
 *
 * Io_Init --
 *
 *      Este procedimiento es llamado siempre que se crea una nueva maquina.
 *      Su trabajo es inicializar la parte de la estructura de la maquina
 *      que se refiere a la entrada/salida.
 *
 * Resultados:
 *      Ninguno.
 *
 * Efectos Laterales:
 *      Depende de los dispositivos de E/S (El que hace la llamada no necesita
 *      conocer esto).
 *
 *----------------------------------------------------------------------
 */

void
Io_Init(machPtr)
    register DLXV *machPtr;             /* Nueva maquina. */
{
    machPtr->ioState.flags = IO_TERM_OUTPUT_READY;

    /*
     * Ordena una llamada retrasada de forma periodica para comprobar la
     * entrada del terminal.
     */

    Sim_CallBack(machPtr, INPUT_WAIT, InputTimerProc, (ClientData) machPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Io_Read --
 *
 *      Lee el registro de un dispositivo de E/S, si existe alguno en la
 *      direccion proporcionada.
 *
 * Resultados:
 *      El valor de retorno es no-cero si hay un registro de E/S en esa
 *      direccion; en otro caso el valor devuelto es cero. Si el registro
 *      existe, su contenido es almacenado en *valuePtr.
 *
 * Efectos Laterales:
 *      Ninguno.
 *
 *----------------------------------------------------------------------
 */

long int
Io_Read(machPtr, address, valuePtr)
    register DLXV *machPtr;     /* Maquina cuya memoria va ser leida. */
    unsigned long int address;  /* Direccion deseada del registro de E/S. */
    long int *valuePtr;         /* Almacena el contenido del registro de
				 * E/S en esta variable. */
{
    long int result = 0;

    switch (address) {

	case IO_RECV_CONTROL:
	    if (machPtr->ioState.flags & IO_TERM_INPUT_READY) {
		result |= IO_READY;
	    }
	    if (machPtr->ioState.flags & IO_TERM_INPUT_IE) {
		result |= IO_IE;
	    }
	    break;

	case IO_RECV_DATA:
	    result = machPtr->ioState.input & 0xff;
	    if (machPtr->ioState.flags & IO_TERM_INPUT_READY) {
		machPtr->ioState.flags &= ~IO_TERM_INPUT_READY;
		if (machPtr->ioState.flags & IO_TERM_INPUT_IE) {
		    Cop0_IntPending(machPtr, cero, muno);
		}
		CheckInput(machPtr);
	    }
	    break;

	case IO_TRANS_CONTROL:
	    if (machPtr->ioState.flags & IO_TERM_OUTPUT_READY) {
		result |= IO_READY;
	    }
	    if (machPtr->ioState.flags & IO_TERM_OUTPUT_IE) {
		result |= IO_IE;
	    }
	    break;

	case IO_TRANS_DATA:
	    break;

	default:
	    return 0;
    }

    *valuePtr = result;
    return 1;
}

/*
 *----------------------------------------------------------------------
 *
 * Io_Write --
 *
 *      Escribe en un registro de un dispositivo de E/S, si es que existe
 *      alguno en la direccion proporcionada.
 *
 * Resultados:
 *      El valor devuelto es no-cero si hay algun registro de E/S en esa
 *      direccion; en otro caso se devuleve cero. Si el registro existe,
 *      su contenido es sobreescrito con el nuevo valor.
 *
 * Efectos Laterales:
 *      La posicion de memoria proporcionada es modificada, lo cual puede
 *      hacer que se ponga en marcha un dispositivo de E/S u otras cosas.
 *
 *----------------------------------------------------------------------
 */

long int
Io_Write(machPtr, address, value, size)
    register DLXV *machPtr;     /* Maquina cuya memoria va a ser escrita. */
    unsigned long int address;  /* Direccion deseada del registro de E/S. */
    long int value;             /* Nuevo valor para el registro de E/S. */
    long int size;              /* Tamao del valor en bytes (1, 2, o 4). */
{
    switch (address & ~0x3) {

	case IO_RECV_CONTROL:
	    if (((address & 0x3) + size) < 4) {
		break;
	    }
	    if (value & IO_IE) {
		if (!(machPtr->ioState.flags & IO_TERM_INPUT_IE)) {
		    machPtr->ioState.flags |= IO_TERM_INPUT_IE;
		    if (machPtr->ioState.flags & IO_TERM_INPUT_READY) {
			Cop0_IntPending(machPtr, cero, uno);
		    }
		}
	    } else {
		if ((machPtr->ioState.flags &
			(IO_TERM_INPUT_IE|IO_TERM_INPUT_READY))
			== (IO_TERM_INPUT_IE|IO_TERM_INPUT_READY)) {
		    Cop0_IntPending(machPtr, cero, muno);
		}
		machPtr->ioState.flags &= ~IO_TERM_INPUT_IE;
	    }
	    break;

	case IO_RECV_DATA:
	    break;

	case IO_TRANS_CONTROL:
	    if (((address & 0x3) + size) < 4) {
		break;
	    }
	    if (value & IO_IE) {
		if (!(machPtr->ioState.flags & IO_TERM_OUTPUT_IE)) {
		    if (machPtr->ioState.flags & IO_TERM_OUTPUT_READY) {
			Cop0_IntPending(machPtr, cero, uno);
		    }
		    machPtr->ioState.flags |= IO_TERM_OUTPUT_IE;
		}
	    } else {
		if ((machPtr->ioState.flags &
			(IO_TERM_OUTPUT_IE|IO_TERM_OUTPUT_READY))
			== (IO_TERM_OUTPUT_IE|IO_TERM_OUTPUT_READY)) {
		    Cop0_IntPending(machPtr, cero, muno);
		}
		machPtr->ioState.flags &= ~IO_TERM_OUTPUT_IE;
	    }
	    break;

	case IO_TRANS_DATA:
	    if (((address & 0x3) + size) < 4) {
		break;
	    }
	    if (machPtr->ioState.flags & IO_TERM_OUTPUT_READY) {
		char c;

		/*
		 * Empieza a trasmitir datos, marca el transmisor como no
		 * preparado,  y programa la maquina para que el transmisor
		 * sea marcado como preparado mas tarde.
		 */

		c = value;
		if (write(1, &c, 1) != 1) {
		    BWCCMessageBox(GetFocus(),"Internal error in stdout!", "Error", MB_OK | MB_ICONEXCLAMATION);
		    exit(1);
		}
		machPtr->ioState.flags &= ~IO_TERM_OUTPUT_READY;
		if (machPtr->ioState.flags & IO_TERM_OUTPUT_IE) {
		    Cop0_IntPending(machPtr, cero, muno);
		}
		Sim_CallBack(machPtr, OUTPUT_DELAY, MarkOutput,
			(ClientData) machPtr);
	    }
	    break;

	default:
	    return 0;
    }
    return 1;
}

/*
 *----------------------------------------------------------------------
 *
 * Io_BeginSim --
 *
 *      Este procedimiento es invocado justo antes de que comience la
 *      simulacion, de manera que todas aquellas cosas relacionadas con
 *      la E/S ( como resetear las caracteristicas del terminal) puedan
 *      ser realizadas.
 *
 * Resultados:
 *      Ninguno.
 *
 * Efectos Laterales:
 *      Ninguno.
 *
 *----------------------------------------------------------------------
 */
#pragma argsused
void
Io_BeginSim(machPtr)
    register DLXV *machPtr;             /* Maquina que se esta simulando. */
{

    /*
     * Salva el estado del terminal, y lo pone en modo raw durante la
     * simulacion.
     */

     _InitEasyWin();
}

/*
 *----------------------------------------------------------------------
 *
 * Io_EndSim --
 *
 *      Este procedimiento es invocado justo despues de que un comando de
 *      simulacion es completado, de manera que las E/S se puedan realizar
 *      ( como por ejemplo resetear las caracteristicas del terminal para
 *      poder procesar los comandos del DLXVsim).
 *
 * Resultados:
 *      Ninguno.
 *
 * Side effects:
 *      Ninguno.
 *
 *----------------------------------------------------------------------
 */

void
Io_EndSim(machPtr)
    register DLXV *machPtr;             /* Maquina simulada. */
{
    /*
     * Lee un caracter pendiente en la entrada, si lo hay, y restaura
     * el estado del terminal.
     */

    CheckInput(machPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * CheckInput --
 *
 *      Comprueba si ha llegado a la entrada algun caracter;  si es asi, lo
 *      lee y actualiza los registros de control de E/S.
 *
 * Resultados:
 *      Ninguno.
 *
 * Efecto Laterales:
 *      El estado del dispositivo de E/S puede cambiar, e interrupciones
 *      pueden ocurrir.
 *
 *----------------------------------------------------------------------
 */

static void
CheckInput(machPtr)
    register DLXV *machPtr;             /* Maquina en la que comprobar
					 * la E/S. */
{
    char c;

    /*
     * Comprueba si hay algun caracter en el terminal.
     */
	if (kbhit())
	  {
	     c=getch();
	     if (c=='s' || c=='S') {raise(SIGINT);}
	     else {
		 machPtr->ioState.input = c;
		 machPtr->ioState.flags |= IO_TERM_INPUT_READY;
		 if (machPtr->ioState.flags & IO_TERM_INPUT_IE) {
		  Cop0_IntPending(machPtr, cero, uno); }
		  }
	  }
}

/*
 *----------------------------------------------------------------------
 *
 * InputTimerProc --
 *
 *      Este procedimiento es invocado periodicamente durante la entrada
 *      para ver si alguna E/S se ha completado.
 *
 * Resultados:
 *      Ninguno.
 *
 * Efectos Laterales:
 *      El estado del dispositivo de E/S puede cambiar, e interrupciones
 *      pueden ocurrir.
 *
 *----------------------------------------------------------------------
 */

static void
InputTimerProc(machPtr)
    register DLXV *machPtr;             /* Maquina en la que comprobar
					 * la E/S. */
{

    CheckInput(machPtr);

    /*
     * Hacemos que se produzca una nueva llamada mas tarde.
     */
    Sim_CallBack(machPtr, INPUT_WAIT, InputTimerProc, (ClientData) machPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * MarkOutput --
 *
 *      Este procedimiento es llamado por el simulador despues de que un
 *      numero determinado de instrucciones han sido ejecutadas desde que
 *      comenzo la operacion de salida.
 *
 * Resultados:
 *      Ninguno.
 *
 * Efectos Laterales:
 *      La unidad de transmision es marcada como "ready" otra vez.
 *
 *----------------------------------------------------------------------
 */

static void
MarkOutput(machPtr)
    DLXV *machPtr;           /* Maquina cuyo transmisor debe ser marcado como
			      * preparado nuevamente. */
{
    machPtr->ioState.flags |= IO_TERM_OUTPUT_READY;
    if (machPtr->ioState.flags & IO_TERM_OUTPUT_IE) {
	Cop0_IntPending(machPtr, cero, uno);
    }
}
