/*
 * sym.c --
 *
 *      Este fichero contiene procedimientos que manipulan los
 *	valores de los simbolos y expresiones que contengan dichos
 *      simbolos.
 *
 * Copyright 1989 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#include <alloc.h>
#include <ctype.h>
#include "hash.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dlx.h"
#include "sym.h"

/*
 * Para cada simbolo definido existe una estructura de la siguiente forma
 * en la tabla de simbolos de la maquina. En ella pueden haber numerosos
 * simbolos con el mismo nombre ( en diferentes ficheros), de manera que
 * habra una lista de estas estructuras colgando de cada entrada de la
 * tabla hash.
 */

typedef struct Sym {
    char *fileName;             /* Nombre del fichero en el cual el simbolo
				 * esta definido. */
    unsigned long int address;  /* Direccion del simbolo, si esta definida. */
    long int flags;             /* Varios flags:  ver abajo. */
    struct Sym *nextPtr;        /* Siguiente simbolo con el mismo nombre, o
				 * NULL para final de lista. */
} Sym;

/*
 * Flags para la estructyra Sym:
 *
 * S_GLOBAL -                   1 significa que es un simbolo global.
 * S_NO_ADDR -                  1 significa que este simbolo no tiene
 *				todavia una direccion definida para el
 *				(En estos casos S_GLOBAL esta siempre
 *				activado).
 * S_REG -                      1 significa que el simbolo es un registro.
 */

#define S_GLOBAL        1L
#define S_NO_ADDR       2L
#define S_REG           4L

/*
 * La siguiente estructura de datos describe el estado del analisis de una
 * expresion.
 * Esta estructura sirve como informacion de intercambio entre las rutinas
 * de este modulo.
 */

typedef struct {
    DLXV *machPtr;              /* Maquina en la que se encuentra la tabla
				 * de simbolos y donde se almacenaran los
				 * mensajes de error. */
    char *fileName;             /* Nombre del fichero asociado al simbolo;
				 * Ver argumento de Sym_GetSym con el mismo
				 * nombre para mas informacion. */
    long int ignoreUndef;       /* No-cero significa que no hay que
				 * preocuparse por los simbolos
				 * indefinidos. */
    char *expr;                 /* Posicion del siguiente caracter que va a
				 * ser examinado en la expresion. */
    long int token;             /* Tipo del ultimo token a ser analizado de
				 * la expresion. Ver abajo para las
				 * definiciones. */
    char *tokenChars;           /* Puntero al primer caracter de la expresion
				 * del token (por ej: el valor de expr antes
				 * de la llamada a Lex que produjo el
				 * token). */
    long int number;            /* Si el token es un numero, proporciona el
				 * valor del numero. */
} ExprInfo;

/*
 * Los tipos de token son definidos abajo. Ademas, hay una tabla asociada
 * a la precedencia de cada operador. El orden de los typos es importante.
 * Consulta el codigo antes de cambiarlos.
 */

#define NUMBER          0L
#define OPEN_PAREN      1L
#define CLOSE_PAREN     2L
#define END             3L

/*
 * Operadores binarios:
 */

#define MULT            8L
#define DIVIDE          9L
#define MOD             10L
#define PLUS            11L
#define MINUS           12L
#define LEFT_SHIFT      13L
#define RIGHT_SHIFT     14L
#define BIT_AND         15L
#define BIT_XOR         16L
#define BIT_OR          17L

/*
 * Operadores unarios:
 */

#define UNARY_MINUS     20L
#define BIT_NOT         21L

/*
 * Tabla de precedencia.  Los valores para los tokens que no son
 * operadores son ignorados.
 */

static long int far exprPrecTable[] = {
    0, 0, 0, 0, 0, 0, 0, 0,
    10, 10, 10,                         /* MULT, DIV, MOD */
    9, 9,                               /* MAS, MENOS */
    8, 8,                               /* DEP_IZQ, DESP_DERECHA */
    5,                                  /* BIT_AND */
    4,                                  /* BIT_XOR */
    3,                                  /* BIT_OR */
    0, 0,
    11, 11                              /* MENOS_UNARIO, BIT_NOT */
};

/*
 *----------------------------------------------------------------------
 *
 * Sym_Tabla
 *	Inicializa la tabla de simbolos
 *
 * Resultados:
 *	Borra todos los simbolos asociados a la maquina simulada
 *
 * Efectos laterales:
 *	La tabla symbols de machPtr es inicializada como si hubiera sido
 * 	creada nueva.
 *
 *-----------------------------------------------------------------------
 */

void Sym_Tabla(machPtr, memSize)
 DLXV *machPtr;
 long int memSize;
 {
  Hash_DeleteTable(&machPtr->symbols); 
  Hash_InitTable(&machPtr->symbols, cero, HASH_STRING_KEYS); 
  Sym_AddSymbol(machPtr, "", "memSize", memSize*4, SYM_GLOBAL);
 }

/*
 *--------------------------------------------------------------------
 *
 * Sym_DeleteSymbols --
 *
 *      Borra todos los simbolos asociados con un determinado fichero de la
 *	tabla de simbolos de la maquina.
 *
 * Resultados:
 *      Ninguno.
 *
 * Efectos Laterales:
 *      Todos los simbolos para "fileName" son borrados de la tabla de
 *	simbolos de "machPtr". Esto es normalmente hecho antes de volver
 *	a cargar un fichero en la memoria.
 *
 *----------------------------------------------------------------------
 */

void
Sym_DeleteSymbols(machPtr, fileName)
    DLXV *machPtr;		/* Maquina a manipular. */
    char *fileName;             /* Nombre del fichero cuyos simbolos deben
				 * ser borrados. */
{
    Hash_Entry *hPtr;
    register Sym *symPtr;
    register Sym *prevPtr;
    Hash_Search search;

    for (hPtr = Hash_EnumFirst(&machPtr->symbols, &search);
	    hPtr != NULL; hPtr = Hash_EnumNext(&search)) {
	for (symPtr = (Sym *) Hash_GetValue(hPtr), prevPtr = NULL;
		symPtr != NULL; prevPtr = symPtr, symPtr = symPtr->nextPtr) {
	    if (strcmp(symPtr->fileName, fileName) != 0) {
		continue;
	    }
	    if (prevPtr == NULL) {
		Hash_SetValue(hPtr, symPtr->nextPtr);
	    } else {
		prevPtr->nextPtr = symPtr->nextPtr;
	    }
	  farfree(symPtr->fileName);
	  farfree((char far *)symPtr); /* ***** Modificado ***** */
	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Sym_AddSymbol --
 *
 *      Aade un nuevo simbolo a la tabla de la maquina.
 *
 * Resultados:
 *      Normalmente se devuelve TCL_OK, y una cadena de error puede ser
 *	incluida en el machPtr si se ha producido un error.
 *
 * Efectos Laterales:
 *      El nombre es introducido en la tabla de simbolos, si aun no esta
 *	alli. Si address es -1 y global es 1, entonces el nombre es
 *	introducido como un simbolo global pero no le es asignada
 *	ninguna direccion ( Esa asignacion se llevara a cabo mas tarde).
 *
 *----------------------------------------------------------------------
 */

long int
Sym_AddSymbol(machPtr, fileName, name, address, flags)
    DLXV *machPtr;              /* Maquina a manipular. */
    char *fileName;             /* Nombre del fichero en el cual el simbolo
				 * esta definido. */
    char *name;                 /* Nombre del simbolo. */
    unsigned long int address;  /* Direccion correspondiente al simbolo, o
				 * SYM_NO_ADDR si el simbolo esta siendo
				 * declarado como global antes de ser
				 * definido. */
    long int flags;             /* Combinacion OR de flags: cualquiera de
				 * los siguientes:
				 * SYM_REGISTER, SYM_GLOBAL, SYM_NO_ADDR. */
{
    register Sym *symPtr;
    Hash_Entry *hPtr;

    hPtr = Hash_CreateEntry(&machPtr->symbols, (Address) name,
	    (Boolead *) NULL);
    symPtr = (Sym *) Hash_GetValue(hPtr);
    if (symPtr != NULL) {
	Sym *globalPtr, *localPtr;

	/*
	 * Vemos si existen otros simbolos que pueden coincidir con este.
	 * Si hay un error debido a un conflicto, Este es manejado mas
	 * adelante.
	 */

	globalPtr = localPtr = NULL;
	for ( ; symPtr != NULL; symPtr = symPtr->nextPtr) {
	    if (strcmp(symPtr->fileName, fileName) == 0) {
		localPtr = symPtr;
	    } else {
		if ((symPtr->flags & (S_GLOBAL|S_NO_ADDR)) == S_GLOBAL) {
		    globalPtr = symPtr;
		}
	    }
	}

	if (localPtr != NULL) {

	    /*
	     * Esta esta llamada unicamente actualizando la informacion
	     * de un simbolo anteriormente introducido? Si es asi lo
	     * hacemos.
	     */

	    if (flags & SYM_GLOBAL) {
		localPtr->flags |= S_GLOBAL;
	    }
	    if (!(flags & SYM_NO_ADDR)) {
		if (!(localPtr->flags & S_NO_ADDR)) {
		    sprintf((char far *)machPtr->result,
			    "Simbol \"%.50s\" undefined in %.50s",
			    name, fileName);
		    return TCL_ERROR;
		}
		localPtr->address = address;
		localPtr->flags &= ~S_NO_ADDR;
	    }

	    /*
	     * Esta este nuevo simbolo en conflicto con otros simbolos
	     * globales definidos?
	     */

	    if ((globalPtr != NULL)
		    && ((localPtr->flags & (S_GLOBAL|S_NO_ADDR)) == S_GLOBAL)) {
		globalConflict:
		sprintf((char far *)machPtr->result,
			"Simbol \"%.50s\" undefined in %.50s",
			name, globalPtr->fileName);
		return TCL_ERROR;
	    }
	    return TCL_OK;
	}

	if ((globalPtr != NULL)
		&& ((flags & (SYM_GLOBAL|SYM_NO_ADDR)) == S_GLOBAL)) {
	    goto globalConflict;
	}
    }

    /*
     * Crea e inicializa un nuevo simbolo.
     */

    symPtr = (Sym far *) farcalloc(1, sizeof(Sym));
    symPtr->fileName = (char far *) farcalloc(1, (unsigned long int) (_fstrlen(fileName) + 1));
    strcpy(symPtr->fileName, fileName);
    symPtr->address = address;
    if (flags & SYM_GLOBAL) {
	symPtr->flags |= S_GLOBAL;
    }
    if (flags & SYM_NO_ADDR) {
	symPtr->flags |= S_NO_ADDR;
    }
    if (flags & SYM_REGISTER) {
	symPtr->flags |= S_REG;
    }
    symPtr->nextPtr = (Sym *) Hash_GetValue(hPtr);
    Hash_SetValue(hPtr, symPtr);

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Sym_GetSym --
 *
 *      Busca un simbolo y devuelve su direccion.
 *
 * Resultados:
 *      El valor devuelto es uno de los siguientes: SYM_FOUND (El caso normal
 *	en el que el simbolo existe), SYM_REGISTER (Si se encontro el simbolo
 *	y es un registro), SYM_NOT_FOUND si no se encontro ese simbolo,
 *	SYM_AMBIGUOUS si el simbolo esta definido varias veces, SYM_REG_NOT_OK
 *	si el nombre se refiere a un registro pero no se permiten registros en
 *	este momento. Si SYM_FOUND o SYM_REGISTER es el valor devuelto *addPtr
 *	es modificado para contener la direccion en memoria del simbolo o el
 *	numero del registro.
 *
 * Efectos Laterales:
 *      Ninguno.
 *
 *----------------------------------------------------------------------
 */

long int
Sym_GetSym(machPtr, fileName, name, flags, addrPtr)
    DLXV *machPtr;              /* Maquina a usar para buscar el simbolo. */
    char *fileName;             /* Si no es null significa que solo hay
				 * que considerar los simbolos que
				 * pertenezcan a ese fichero.
				 * NULL significa que hemos de considerar
				 * cualquier simbolo global o que no sea
				 * ambiguo. */
    char *name;                 /* Nombre del simbolo deseado. */
    long int flags;             /* Cero o uno de los siguientes:
				 * SYM_REGS_OK considera registeros normales;
				 * SYM_PSEUDO_OK acepta nombres de pseudo
				 * registros (por ej. vlr). */
    unsigned long int *addrPtr; /* La direccion del simbolo (o numero del
				 * registro) es almacenado aqui. */
{
    Hash_Entry *hPtr;
    register Sym *symPtr;
    Sym *matchPtr;
    long int matchCount;

    matchPtr = NULL;
    matchCount = 0;
    hPtr = Hash_FindEntry(&machPtr->symbols, name);
    if (hPtr != NULL) {

	/*
	 * Recorre todos los simbolos con ese nombre.
	 */

	for (symPtr = (Sym *) Hash_GetValue(hPtr); symPtr != NULL;
		symPtr = symPtr->nextPtr) {
	    if (symPtr->address == SYM_NO_ADDR) {
		continue;
	    }
	    if (symPtr->flags & S_GLOBAL) {
		found:
		*addrPtr = symPtr->address;
		if (symPtr->flags & S_REG) {
		    if (flags & (SYM_REGS_OK|SYM_PSEUDO_OK)) {
			return SYM_REGISTER;
		    }
		    return SYM_REG_NOT_OK;
		}
		return SYM_FOUND;
	    }
	    if (fileName != NULL) {
		if (strcmp(fileName, symPtr->fileName) == 0) {
		    goto found;
		}
	    } else {
		matchPtr = symPtr;
		matchCount++;
	    }
	}
    }

    /*
     * Si no se ha encontrado, vemos si el nombre se refiere a un
     * nombre de registro.
     */

    if (matchCount == 0) {
	long int num;
	char *end;

	if (*name == 'v') {
	    num = strtoul(name+1, &end, 10);
	    if ((end != (name+1)) && (*end == 0) && (num<= 7) && (num >= 0)) {
		if (num >= 8) {
		    return SYM_NOT_FOUND;
		}
		*addrPtr = num;
		if (!(flags & SYM_VREGS_OK)) {
		    return SYM_VREG_FOUND;
		}
		return SYM_REGISTER;
	    }
	}
	if (*name == 'f') {
	    num = strtoul(name+1, &end, 10);
	    if ((end != (name+1)) && (*end == 0) && (num <= 31) &&(num >= 0)) {
		if (num >= 32) {
		    return SYM_NOT_FOUND;
		}
		*addrPtr = num;
		if (!(flags & SYM_FREGS_OK)) {
		    return SYM_FREG_FOUND;
		}
		return SYM_REGISTER;
	    }
	}
	if ((*name == 'r') || (*name == '$')) {
	    num = strtoul(name+1, &end, 10);
	    if ((end != (name+1)) && (*end == 0) && (num <= 31) && (num >= 0)) {
		gotReg:
		if (num >= 32) {
		    if (!(flags & SYM_PSEUDO_OK)) {
			return SYM_NOT_FOUND;
		    }
		} else if (!(flags & (SYM_PSEUDO_OK|SYM_REGS_OK))) {
		    return SYM_REG_NOT_OK;
		}
		*addrPtr = num;
		return SYM_REGISTER;
	    }
	}
	if ((flags & SYM_PSEUDO_OK) && (strcmp(name, "pc") == 0)) {
	    num = PC_REG; goto gotReg;
	} else if ((flags & SYM_PSEUDO_OK) && (strcmp(name, "npc") == 0)) {
	    num = NEXT_PC_REG; goto gotReg;
	} else if ((flags & SYM_PSEUDO_OK) && (strcmp(name, "vlr") == 0)) {
	    num = NEXT_PC_REG+1; goto gotReg;
	} else if ((flags & SYM_PSEUDO_OK) && (strcmp(name, "vm") == 0)) {
	    num = NEXT_PC_REG+2; goto gotReg;
	} else if ((flags & SYM_PSEUDO_OK) && (strcmp(name, "vmlo") == 0)) {
	    num = NEXT_PC_REG+3; goto gotReg;
	} else if (strcmp(name, "at") == 0) {
	    num = 1; goto gotReg;
	}
    }

    /*
     * Una ultima comprobacion:  Si hay algun simbolo local no ambiguo,y eso
     * esta permitido, entonces lo devuelve.
     */

    if ((fileName == NULL) && (matchCount == 1)) {
	symPtr = matchPtr;
	goto found;
    }
    if (matchCount > 1) {
	return SYM_AMBIGUOUS;
    }
    return SYM_NOT_FOUND;
}

/*
 *----------------------------------------------------------------------
 *
 * Lex --
 *
 *      Analizador lexico para analizar expresiones.
 *
 * Resultados:
 *      Se devuelve TCL_OK a menos que ocurra un error al hacer el analisis
 *	lexico. En ese caso un resultado de error es devuelto, usando
 *	machPtr->result para alamacenar el mensaje de error. En el caso de
 *	no ocurrir errores, el token y (posiblemente) loas campos numericos
 *	de infoPtr son actualizados para referirse al siguiente simbolo de
 *	la expresion almacenada en el string, haciendo ademas avanzar el
 *	campo expr.
 *
 * Efectos Laterales:
 *      Ninguno.
 *
 *----------------------------------------------------------------------
 */

static long int
Lex(infoPtr)
    register ExprInfo *infoPtr;  /* Describe el estado del analisis. */
{
    register char *p, c;

    /*
     * El siguiente token puede ser:
     * (a)      Un numero (convertirlo de ASCII a binario).
     * (b)      El nombre de un simbolo (buscar el simbolo).
     * (c)      Espacio (Saltarselo).
     * (d)      Un operador (Ver que clase de operador es).
     */

    p = infoPtr->expr;
    infoPtr->tokenChars = p;
    c = *p;
    while (isspace(c)) {
	p++;  c = *p;
    }
    infoPtr->expr = p;
    if (!isascii(c)) {
	infoPtr->token = END;
	return TCL_OK;
    }
    if (c == '#') {
	if (((*(p+1) == '-') && (!isdigit(*(p+2)))) ||
	    (!isdigit(*(p+1)) && (*(p+1) != '-'))) {
	    return TCL_ERROR;
	}
	infoPtr->token = NUMBER;
	infoPtr->number = strtol(p+1, &infoPtr->expr, 0);
	return TCL_OK;
    }
    if (isdigit(c)) {
	infoPtr->token = NUMBER;
	infoPtr->number = strtol(p, &infoPtr->expr, 0);
	return TCL_OK;
    }
    if (isalpha(c) || (c == '_')) {
	char savedChar;
	long int result;

	for (p++; isalnum(*p) || (*p == '_') || (*p == '$'); p++) {
	    /* Cuerpo vacio;  Salta hasta el final del simbolo. */
	}
	savedChar = *p;
	*p = 0;
	result = Sym_GetSym(infoPtr->machPtr, infoPtr->fileName,
		infoPtr->expr, cero, (unsigned long int *) &infoPtr->number);
	if (result == SYM_NOT_FOUND) {
	    if (infoPtr->ignoreUndef) {
		infoPtr->number = 0;
	    } else {
		sprintf((char far *)infoPtr->machPtr->result,
			"Simbol undefined \"%.50s\"", infoPtr->expr);
		*p = savedChar;
		return TCL_ERROR;
	    }
	} else if (result == SYM_AMBIGUOUS) {
	    sprintf((char far *)infoPtr->machPtr->result,
		    "Simbol \"%.50s\" defined many times", infoPtr->expr);
	    *p = savedChar;
	    return TCL_ERROR;
	} else if (result == SYM_REG_NOT_OK) {
	    sprintf((char far *)infoPtr->machPtr->result,
		    "Cannot use register name in expresion");
	    *p = savedChar;
	    return TCL_ERROR;
	}
	infoPtr->token = NUMBER;
	*p = savedChar;
	infoPtr->expr = p;
	return TCL_OK;
    }

    infoPtr->expr = p+1;
    switch (c) {
	case '(':
	    infoPtr->token = OPEN_PAREN;
	    return TCL_OK;

	case ')':
	    infoPtr->token = CLOSE_PAREN;
	    return TCL_OK;

	case '*':
	    infoPtr->token = MULT;
	    return TCL_OK;

	case '/':
	    infoPtr->token = DIVIDE;
	    return TCL_OK;

	case '%':
	    infoPtr->token = MOD;
	    return TCL_OK;

	case '+':
	    infoPtr->token = PLUS;
	    return TCL_OK;

	case '-':
	    infoPtr->token = MINUS;
	    return TCL_OK;

	case '<':
	    if (p[1] == '<') {
		infoPtr->expr = p+2;
		infoPtr->token = LEFT_SHIFT;
	    } else {
		infoPtr->token = END;
	    }
	    return TCL_OK;

	case '>':
	    if (p[1] == '>') {
		infoPtr->expr = p+2;
		infoPtr->token = RIGHT_SHIFT;
	    } else {
		infoPtr->token = END;
	    }
	    return TCL_OK;

	case '&':
	    infoPtr->token = BIT_AND;
	    return TCL_OK;

	case '^':
	    infoPtr->token = BIT_XOR;
	    return TCL_OK;

	case '|':
	    infoPtr->token = BIT_OR;
	    return TCL_OK;

	case '~':
	    infoPtr->token = BIT_NOT;
	    return TCL_OK;

	case 0:
	default:
	    infoPtr->token = END;
	    infoPtr->expr = p;
	    return TCL_OK;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * GetValue --
 *
 *      Analiza un valor de lo que queda de la expresion en infoPtr.
 *
 * Resultados:
 *      Normalmente se devuelve TCL_OK.  El valor del numero analizado
 *	es devuelto en infoPtr->number. Si ocurre un error, entonces
 *	machPtr->result contiene un mensaje de error y se devuelve
 *	TCL_ERROR.
 *
 * Efectos Laterales:
 *      La informacion que queda en la expresion es analizada, y los
 *	campos expr, token y tokenChars de infoPtr son actualizados.
 *      La informacion es analizada hasta que se llega al final de la
 *	expresion (caracter null o cerrar parentesis), ocurre un error,
 *	o se encuentra un operador binario con precedencia <= que prec.
 *	En cualquiera de estos casos, infoPtr->token quedara apuntando
 *	al token que esta situado despues de la expresion.
 *
 *----------------------------------------------------------------------
 */

static long int
GetValue(infoPtr, prec)
    register ExprInfo *infoPtr; /* Describe el estado del analisis justo antes
				 * del valor (por ej. Lex sera llamado para
				 * obtener el primer token del valor). */
    long int prec;              /* Trata cualquier operador sin parentesis
				 * con precedencia <=  este como el final
				 * de la expresion. */
{
    long int result, operator, operand;
    long int gotOp;             /* Un valor no cero significa que aun se
				 * esta analizando el operador. No llamar a
				 * al lex otra vez. */
    char *savedPtr;

    /*
     * Hay dos fases en este procedimiento.  Primero, toma un valor inicial.
     * Entonces, analiza pares (operador binario, valor) hasta finalizar.
     */

    gotOp = 0;
    result = Lex(infoPtr);
    if (result != TCL_OK) {
	return result;
    }
    if (infoPtr->token == OPEN_PAREN) {

	/*
	 * Subexpresion entre parentesis.
	 */

	result = GetValue(infoPtr, muno);
	if (result != TCL_OK) {
	    return result;
	}
	if (infoPtr->token != CLOSE_PAREN) {
	    Tcl_Return(infoPtr->machPtr, (char far *) NULL, TCL_STATIC);
	    sprintf((char far *)infoPtr->machPtr->result,
		    "Parenthesis not in pair");
	    return TCL_ERROR;
	}
    } else {
	if (infoPtr->token == MINUS) {
	    infoPtr->token = UNARY_MINUS;
	}
	if (infoPtr->token >= UNARY_MINUS) {

	    /*
	     * Procesa operadore unarios.
	     */

	    operator = infoPtr->token;
	    savedPtr = infoPtr->expr;
	    result = GetValue(infoPtr, exprPrecTable[(unsigned int)infoPtr->token]);
	    if (result != TCL_OK) {
		return result;
	    }
	    if (infoPtr->tokenChars == savedPtr) {
		sprintf((char far *)infoPtr->machPtr->result,
			"Expresion fault, Value before operator.");
		return TCL_ERROR;
	    }
	    switch (operator) {
		case UNARY_MINUS:
		    infoPtr->number = -infoPtr->number;
		    break;
		case BIT_NOT:
		    infoPtr->number = ~infoPtr->number;
		    break;
	    }
	    gotOp = 1;
	} else if (infoPtr->token != NUMBER) {
	    return TCL_OK;
	}
    }

    /*
     * Conseguimos el primer operando.  Ahora busca pares (operador, operando).
     */

    if (!gotOp) {
	result = Lex(infoPtr);
	if (result != TCL_OK) {
	    return result;
	}
    }
    while (1) {
	operand = infoPtr->number;
	operator = infoPtr->token;
	if ((operator < MULT) || (operator >= UNARY_MINUS)) {
	    return TCL_OK;
	}
	if (exprPrecTable[(unsigned int)operator] <= prec) {
	    return TCL_OK;
	}

	savedPtr = infoPtr->expr;
	result = GetValue(infoPtr, exprPrecTable[(unsigned int)operator]);
	if (infoPtr->tokenChars == savedPtr) {
	    sprintf((char far *)infoPtr->machPtr->result,
		    "Expresion fault, Value before operator.");
	    return TCL_ERROR;
	}
	if (result != TCL_OK) {
	    return result;
	}
	switch (operator) {
	    case MULT:
		infoPtr->number = operand * infoPtr->number;
		break;
	    case DIVIDE:
		if (infoPtr->number == 0) {
		    Tcl_Return(infoPtr->machPtr,
			    "Division by zero", TCL_STATIC);
		    return TCL_ERROR;
		}
		infoPtr->number = operand / infoPtr->number;
		break;
	    case MOD:
		if (infoPtr->number == 0) {
		    Tcl_Return(infoPtr->machPtr,
			    "Division by zero", TCL_STATIC);
		    return TCL_ERROR;
		}
		infoPtr->number = operand % infoPtr->number;
		break;
	    case PLUS:
		infoPtr->number = operand + infoPtr->number;
		break;
	    case MINUS:
		infoPtr->number = operand - infoPtr->number;
		break;
	    case LEFT_SHIFT:
		infoPtr->number = operand << infoPtr->number;
		break;
	    case RIGHT_SHIFT:
		infoPtr->number = operand >> infoPtr->number;
		break;
	    case BIT_AND:
		infoPtr->number = operand & infoPtr->number;
		break;
	    case BIT_XOR:
		infoPtr->number = operand ^ infoPtr->number;
		break;
	    case BIT_OR:
		infoPtr->number = operand | infoPtr->number;
		break;
	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Sym_EvalExpr --
 *
 *      Dada una expresion incluyendo nombres de simbolos y constantes,
 *	calcula el valor de la expresion.
 *
 * Resultados:
 *      MachPtr->result puede obtener un mensaje de resultado. *valuePtr es
 *	modificado para que contenga el valor de la expresion, y *endPtr es
 *	modificado para que apunte al caracter que termina la expresion.
 *	Si ocurre un error, *endPtr se pone a exprString.
 *
 * Efectos Laterales:
 *      Ninguno.
 *
 *----------------------------------------------------------------------
 */

long int
Sym_EvalExpr(machPtr, fileName, exprString, ignoreUndef, valuePtr, endPtr)
    DLXV *machPtr;              /* Maquina a manipular. */
    char *fileName;             /* Nombre del fichero a usar para buscar
				 * el simbolo. Ver declaracion de los
				 * argumentos de Sym_GetSym para mayor
				 * informacion. */
    char *exprString;           /* Expresion a evaluar. */
    long int ignoreUndef;       /* No cero significa que hemos de tratar los
				 * simbolos indefinidos como si tuvieran la
				 * direccion 0. */
    long int *valuePtr;         /* Almacena el valor de la expresion aqui. */
    char **endPtr;              /* Almacena la direccion del ultimo caracter
				 * aqui. */
{
    ExprInfo info;
    long int result;

    info.machPtr = machPtr;
    info.fileName = fileName;
    info.ignoreUndef = ignoreUndef;
    info.expr = exprString;
    result = GetValue(&info, muno);
    if (result != TCL_OK) {
	*valuePtr = 0;
	*endPtr = exprString;
	return result;
    }
    *valuePtr = info.number;

    /*
     * Si la expresion no contenia informacion, entonces se produce
     * un error.
     */

    if (info.tokenChars == exprString) {
	sprintf((char far *)machPtr->result, "Expresion error");
	return TCL_ERROR;
    }

    /*
     * Salta los caracteres en blancos que existen al final de la expresion.
     */

    while (isspace(*(info.tokenChars))) {
	info.tokenChars++;
    }
    *endPtr = info.tokenChars;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Sym_GetString --
 *
 *      Dada una direccion, devuelve un string describiendo la direccion.
 *
 * Resultados:
 *      El valor devuelto es una descripcion textual de la direccion. Si
 *	es posible, la direccion es identificada simbolicamente (o con
 *	un offset). En el peor de los caos, simplemente sera identificada
 *	como una direccion en hexadecimal.
 *
 * Efectos Laterales:
 *      Ninguno.
 *
 *----------------------------------------------------------------------
 */

char *
Sym_GetString(machPtr, address)
    DLXV *machPtr;              /* Maquina a manipular. */
    unsigned long int address;  /* Direccion para la cual se pide que se
				 * devuelva un string. */
{
    Hash_Search search;
    Hash_Entry *hPtr;
    register Sym *symPtr;
    long int distance, closestDistance;
    char *closestName;
    static char string[100];

    closestDistance = 10000;
    closestName = NULL;

    for (hPtr = Hash_EnumFirst(&machPtr->symbols, &search);
	    hPtr != NULL; hPtr = Hash_EnumNext(&search)) {
	for (symPtr = (Sym *) Hash_GetValue(hPtr); symPtr != NULL;
		symPtr = symPtr->nextPtr) {
	    distance = address - symPtr->address;
	    if ((distance < 0) || (distance >= closestDistance)) {
		continue;
	    }
	    if (distance == 0) {
		return hPtr->key.name;
	    }
	    closestDistance = distance;
	    closestName = hPtr->key.name;
	}
    }
    if (closestName == NULL) {
	sprintf(string, "0x%lx", address);
    } else {
	sprintf(string, "%.50s+0x%lx", closestName, closestDistance);
    }
    return string;
}
