/*
 * list.h --
 *
 * Estructuras, macros y rutinas exportadas por el modulo List.
 *
 * Copyright (C) 1985, 1988 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#ifndef _LIST
#define _LIST

#ifndef _SPRITE
#include "sprite.h"
#endif

/*
 * Este modulo define la abstraccion de una lista, la cual permite unir
 * entre si estructuras de datos diferentes. Las listas estan doblemente
 * enlazadas y son scirculares. Una lista contiene una cabecera seguida por
 * los elementos que la componen, si los hay. ( Una lista vacia, por ello,
 * consiste en un unico elemento, la cabecera, cuyos campos nextPtr y prevPtr
 * apuntan a si mismo). Para referenciar una lista como un todo, el usuario
 * mantiene un puntero a la cabecera; Esa cabecera es inicializada mediante una
 * llamada a List_Init(), la cual crea una lista vacia proporcionando un
 * puntero a una estructura List_Links (descrita abajo).
 *
 * Los enlaces estan presentes en una estructura de dos elementos llamada
 * List_Links. Una lista une registros del tipo List_Links (es decir, cada
 * estructura List_Links apunta a otra estructura List_Links), pero si el
 * List_Links es el primer campo de una estructura mas grande, entonces esas
 * estructuras son unidas entre si de la siguiente manera:
 *
 *	    Cabecera
 *	  (List_Links)		   1er elem.		   2 elemento
 *	-----------------	-----------------	-----------------
 * ..->	|    nextPtr	| ---->	|  List_Links	| ---->	|  List_Links	|----..
 *	| - - - - - - -	|	|		|	|		|
 * ..--	|    prevPtr	| <----	|		| <----	|		|<---..
 *	-----------------	- ---  ---  ---	-	- ---  ---  ---	-
 *				|  resto de la	|	|  resto de la	|
 *				|  estructura	|	|  estructura	|
 *				|		|	|		|
 *				|      ...	|	|      ...	|
 *				-----------------	-----------------
 *
 * Es posible enlazar estructuras a traves de los campos de List_Links los
 * cuales no se encuentran al comienzo del la estructura asociada, pero
 * en este caso es necesario realizar operaciones sobre punteros para
 * encontrar el principio de la estructura mas grande, proporcionando un
 * puntero a algun punto en el interior de la misma.
 *
 * Una estructura tipica podria ser algo como:
 *
 *      typedef struct {
 *                  List_Links links;
 *                  char ch;
 *                  integer flags;
 *      } EditChar;
 *
 * Antes de que un elemento pueda ser insertado en la lista por primera vez,
 * debe ser inicializado realizando una llamada a la macro List_InitElement().
 */


/*
 * Estructura de datos para las listas
 */

typedef struct List_Links {
    struct List_Links *prevPtr;
    struct List_Links *nextPtr;
} List_Links;

/*
 * procedimientos
 */

void	List_Init();    /* Inicializa una cabecera de una lista */
void    List_Insert();  /* Inserta un elemento en la lista */
void    List_ListInsert();  /* Inserta una lista en otra lista */
void 	List_Remove();  /* Borra un elemento de una lista */
void 	List_Move();    /* Mueve un elemento a algun otro lugar de la lista */

/*
 * ----------------------------------------------------------------------------
 *
 * List_InitElement --
 *
 *      Inicializa un elemento de la lista.  Debe ser llamada antes de insertar,
 *	por primera vez, un elemento en una lista.
 *
 * ----------------------------------------------------------------------------
 */
#define List_InitElement(elementPtr) \
    (elementPtr)->prevPtr = (List_Links *) NIL; \
    (elementPtr)->nextPtr = (List_Links *) NIL;

/*
 * Macros para desplazarse paso a paso a traves de determinadas partes de
 * las listas.
 */

/*
 * ----------------------------------------------------------------------------
 *
 * LIST_FORALL --
 *
 *      Macro para realizar una determinada operacion sobre todos los elementos
 *	de una lista.
 *
 *      Uso: LIST_FORALL(headerPtr, itemPtr) {
 *                 / *
 *                   * Operacion sobre itemPtr, el cual apunta a sucesivos
 *		     * miembros de la lista
 *                   *
 *                   * Se puede hacer la primera asignacion
 *                   *          foobarPtr = (Foobar *) itemPtr;
 *                   * para referirnos a toda la estructura Foobar.
 *                   * /
 *             }
 *
 *      Nota: itemPtr debe ser una variable del tipo List_Links, y headerPtr
 *      debe ser evaluado como un puntero a una estructura List_Links.
 *
 * ----------------------------------------------------------------------------
 */

#define LIST_FORALL(headerPtr, itemPtr) \
	for (itemPtr = List_First(headerPtr); \
	     !List_IsAtEnd((headerPtr),itemPtr); \
	     itemPtr = List_Next(itemPtr))

/*
 * ----------------------------------------------------------------------------
 *
 * List_IsEmpty --
 *
 *      Macro: Valor booleano, TRUE si la lista proporcionada no contiene
 *	ningun elemento.
 *
 *      Uso: if (List_IsEmpty(headerPtr)) ...
 *
 * ----------------------------------------------------------------------------
 */

#define List_IsEmpty(headerPtr) \
	((headerPtr) == (headerPtr)->nextPtr)

/*
 * ----------------------------------------------------------------------------
 *
 * List_IsAtEnd --
 *
 *      Macro: Valor booleano, TRUE si itemPtr esta posicionado despues del
 *	final de headerPtr
 *      (por ej., itemPtr es la cabecera de la lista).
 *
 *      Uso: if (List_IsAtEnd(headerPtr, itemPtr)) ...
 *
 * ----------------------------------------------------------------------------
 */


#define List_IsAtEnd(headerPtr, itemPtr) \
        ((itemPtr) == (headerPtr))


/*
 * ----------------------------------------------------------------------------
 *
 * List_First --
 *
 *      Macro que devuelve el primer elemento de la lista, el cual sera la
 *	cabecera si la lista esta vacia.
 *
 *      Uso: firstPtr = List_First(headerPtr);
 *
 * ----------------------------------------------------------------------------
 */

#define List_First(headerPtr) ((headerPtr)->nextPtr)

/*
 * ----------------------------------------------------------------------------
 *
 * List_Last --
 *
 *      Macro que devuelve el ultimo elemento de la lista, el cual es la
 *	cabecera si la lista esta vacia.
 *
 *      Uso: lastPtr = List_Last(headerPtr);
 *
 * ----------------------------------------------------------------------------
 */

#define List_Last(headerPtr) ((headerPtr)->prevPtr)

/*
 * ----------------------------------------------------------------------------
 *
 * List_Prev --
 *
 *      Macro que devuelve el elemento que precede en la lista al elemento
 *	proporcionado. Si el elemnto que se proporciona es el primer elemento
 *	de la lista, List_Prev devuelve la cabecera de la lista.
 *
 *      Uso: prevPtr = List_Prev(itemPtr);
 *
 * ----------------------------------------------------------------------------
 */

#define List_Prev(itemPtr) ((itemPtr)->prevPtr)

/*
 * ----------------------------------------------------------------------------
 *
 * List_Next --
 *
 *      Macro que devuelve el siguiente elemento al proporcionado.
 *      Si el elemento proporcionado es el ultimo elemento de la lista,
 *	List_Next devuelve la cabecera de la lista.
 *
 *      Uso: nextPtr = List_Next(itemPtr);
 *
 * ----------------------------------------------------------------------------
 */

#define List_Next(itemPtr) ((itemPtr)->nextPtr)


/*
 * ----------------------------------------------------------------------------
 *      El procedimiento List_Insert toma dos argumentos.  El primer argumento
 *	es un puntero a la estructura que va a ser insertada en la lista, y el
 *	segundo elemento es un puntero al miembro de la lista despues del
 *	cual el nuevo elemento va a ser insertado. Los Macros son usados para
 *	determinar que miembro ya existente precedera al nuevo.
 *
 *      El procedimiento List_Move toma la misma semantica que List_Insert
 *	para los argumentos.
 *
 *      Los siguientes macros definen donde insertar el nuevo elemento dentro
 *	de la lista:
 *
 *      LIST_AFTER(itemPtr)     --      inserta despues de itemPtr
 *      LIST_BEFORE(itemPtr)    --      inserta antes de itemPtr
 *      LIST_ATFRONT(headerPtr) --      inserta al principio de la lista
 *      LIST_ATREAR(headerPtr)  --      inserta al final de la lista
 *
 *      Por ejemplo,
 *
 *              List_Insert(itemPtr, LIST_AFTER(otherPtr));
 *
 *      insertara itemPtr a continuacion de otherPtr en la lista que contiene a
 *	otherPtr.
 * ----------------------------------------------------------------------------
 */

#define LIST_AFTER(itemPtr) ((List_Links *) itemPtr)

#define LIST_BEFORE(itemPtr) (((List_Links *) itemPtr)->prevPtr)

#define LIST_ATFRONT(headerPtr) ((List_Links *) headerPtr)

#define LIST_ATREAR(headerPtr) (((List_Links *) headerPtr)->prevPtr)

#endif /* _LIST */
